#include <linux/bitops.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/usb.h>
#include <linux/vmalloc.h>
#include <linux/moduleparam.h>
#include <linux/mutex.h>
#include <sound/core.h>
#include <sound/info.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/initval.h>
#include "visualpresenteraudio.h"

static DEFINE_MUTEX(register_mutex);

static void usb_audio_irq(struct urb *urb)
{
	struct vpaudio_dev *vpdev = urb->context;
	struct snd_pcm_substream *substream = vpdev->pcm_substream;
	struct snd_pcm_runtime *runtime = substream->runtime;
	int status, oldptr, stride, len, period_elapsed = 0;

	stride = runtime->frame_bits >> 3;
	
	switch (urb->status) {
		case 0:	{
			if (0x55 == vpdev->data[62] && 0xAA == vpdev->data[63]) {
					goto submit;
			}
			else {
				spin_lock(&vpdev->lock);
				oldptr = vpdev->hwptr_done;
				len = urb->actual_length/stride;
				vpdev->hwptr_done += len;
				if (vpdev->hwptr_done >= runtime->buffer_size) {
					vpdev->hwptr_done -= runtime->buffer_size;
				}
				vpdev->transfer_done += len;
				if (vpdev->transfer_done >= runtime->period_size) {
					vpdev->transfer_done -= runtime->period_size;
					period_elapsed = 1;
				}
				spin_unlock(&vpdev->lock);

				if (oldptr+len > runtime->buffer_size) {
					unsigned int cnt  = runtime->buffer_size-oldptr;
					unsigned int blen = cnt*stride;
					memcpy(runtime->dma_area+oldptr*stride, vpdev->data, blen);
					memcpy(runtime->dma_area, vpdev->data+blen, len*stride-blen);
				} else {
					memcpy(runtime->dma_area + oldptr*stride, vpdev->data, len*stride);
				}

				if (period_elapsed) {
					snd_pcm_period_elapsed(vpdev->pcm_substream);
				}
			}
submit:
			status = usb_submit_urb(vpdev->irq, GFP_ATOMIC);
			if (status) {
				DBG("irq can't resubmit intr.\n");
				goto submit;
			}
			/*else {
				DBG("send urb ok.\n");
			}*/
		}
		break;
		default:
		break;
	}

	return;
}

static struct snd_pcm_hardware vpaudio_hardware = {
	.info 		=	SNDRV_PCM_INFO_MMAP |
				SNDRV_PCM_INFO_MMAP_VALID |
				SNDRV_PCM_INFO_INTERLEAVED |
				SNDRV_PCM_INFO_HALF_DUPLEX,
				
	.formats 	  = 	SNDRV_PCM_FMTBIT_S16_LE,
	.rates 		  = 	SNDRV_PCM_RATE_22050,
	.rate_min	  = 	22050,
	.rate_max	  = 	22050,
	.channels_min 	  = 	1,
	.channels_max 	  = 	1,		
	.buffer_bytes_max =	64 * 200,
	.period_bytes_min =	64,
	.period_bytes_max =	64 * 200,
	.periods_min 	  =	1,
	.periods_max 	  =	200,
};

static int vpaudio_alloc_vmalloc_buffer(struct snd_pcm_substream *substream, size_t size)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	
	if (runtime->dma_area) {
		if (runtime->dma_bytes >= size) {
			return 0;
		}
		vfree(runtime->dma_area);
	}
	
	runtime->dma_area = vmalloc(size);
	memset(runtime->dma_area, 0, size);
	if (!runtime->dma_area) {
		return -ENOMEM;
	}
	
	runtime->dma_bytes = size;
	
	return 0;
}

static int vpaudio_free_vmalloc_buffer(struct snd_pcm_substream *substream)
{
	struct snd_pcm_runtime *runtime = substream->runtime;

	vfree(runtime->dma_area);
	runtime->dma_area = NULL;
	
	return 0;
}

static int vpaudio_capture_open(struct snd_pcm_substream *substream)
{
	struct vpaudio_dev	*vpdev 	 = substream->private_data;
	struct snd_pcm_runtime 	*runtime = substream->runtime;

	runtime->hw = vpaudio_hardware;
	vpdev->pcm_substream = substream;
	
	DBG("vpaudio_capture_open.\n");

	return 0;
}

static int vpaudio_capture_close(struct snd_pcm_substream *substream)
{
	DBG("vpaudio_capture_close.\n");
	
	return 0;
}

static int vpaudio_capture_hw_params(struct snd_pcm_substream *substream,
			     struct snd_pcm_hw_params *hw_params)
{	
	//struct snd_pcm_runtime *runtime = substream->runtime;
	int ret, size;

	DBG("vpaudio_capture_hw_params.\n");
	size = params_buffer_bytes(hw_params);
	DBG("buffer size = %d.\n", size);
	ret = vpaudio_alloc_vmalloc_buffer(substream, size);
	if (ret < 0) {
		DBG("vpaudio_capture_hw_params can not alloc memory!\n");
		return ret;
	}
	DBG("vpaudio_capture_hw_params entry.\n");

	return 0;
}

static int vpaudio_capture_hw_free(struct snd_pcm_substream *substream)
{
	//struct snd_pcm_runtime *runtime = substream->runtime;
	int	ret;
	
	ret = vpaudio_free_vmalloc_buffer(substream);
	if (ret < 0) {
		DBG("vpaudio_capture_hw_free can not free memory.\n");
	}

	DBG("vpaudio_capture_hw_free.\n");

	return 0;
}

static int vpaudio_capture_prepare(struct snd_pcm_substream *substream)
{
	struct snd_pcm_runtime 	*runtime = substream->runtime;
	struct vpaudio_dev	*vpdev 	 = substream->private_data;

	DBG("vpaudio_capture_prepare.\n");

	spin_lock(&vpdev->lock);
	vpdev->hwptr_done = 0;
	vpdev->transfer_done = 0;

	if (vpdev->start) {
		usb_unlink_urb(vpdev->irq);
		DBG("usb unlink usb.\n");
	}

	spin_unlock(&vpdev->lock);

	DBG("vpaudio_playback_prepare.\n");

	return 0;
}

static int vpaudio_capture_trigger(struct snd_pcm_substream *substream, int cmd)
{
	struct  vpaudio_dev *vpdev = substream->private_data;
	int	status;

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START: {
			DBG("capture : vpaudio_trigger start.\n");
			spin_lock(&vpdev->lock);
			if (0 == vpdev->start) {
				vpdev->start = 1;
			}
			status = usb_submit_urb(vpdev->irq, GFP_KERNEL);
			if (status) {
				DBG("irq can't resubmit intr.\n");
			}
			else {
				DBG("send urb ok.\n");
			}
			spin_unlock(&vpdev->lock);
		}
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		DBG("capture : vpaudio_trigger stop.\n");
		break;
	case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
		DBG("capture : vpaudio_trigger push.\n");
		break;
	case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
		DBG("capture : vpaudio_trigger release.\n");
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static snd_pcm_uframes_t vpaudio_capture_pointer(struct snd_pcm_substream *substream)
{
	snd_pcm_uframes_t hwptr_done;
	//struct snd_pcm_runtime 	*runtime = substream->runtime;
	struct vpaudio_dev	*vpdev 	 = substream->private_data;

	spin_lock(&vpdev->lock);
	hwptr_done = vpdev->hwptr_done;
	//DBG("hw ptr = %d.\n", hwptr_done);
	spin_unlock(&vpdev->lock);

	return hwptr_done;
}

static int vpaudio_playback_open(struct snd_pcm_substream *substream)
{
	struct vpaudio_dev	*vpdev 	 = substream->private_data;
	struct snd_pcm_runtime 	*runtime = substream->runtime;
	
	DBG("vpaudio_playback_open.\n");
	runtime->hw = vpaudio_hardware;
	vpdev->pcm_substream = substream;

	return 0;
}

static int vpaudio_playback_close(struct snd_pcm_substream *substream)
{
	DBG("vpaudio_playback_close.\n");
	
	return 0;
}

static int vpaudio_playback_hw_params(struct snd_pcm_substream *substream,
			     struct snd_pcm_hw_params *hw_params)
{
	//struct snd_pcm_runtime *runtime = substream->runtime;
	int ret, size;
	DBG("vpaudio_playback_hw_params.\n");

	size = params_buffer_bytes(hw_params);
	ret = vpaudio_alloc_vmalloc_buffer(substream, size);
	if (ret < 0) {
		DBG("vpaudio_playback_hw_params can not alloc memory!\n");
		return ret;
	}

	return 0;
}

static int vpaudio_playback_hw_free(struct snd_pcm_substream *substream)
{
	//struct snd_pcm_runtime *runtime = substream->runtime;
	int	ret;
	
	DBG("vpaudio_playback_hw_free.\n");
	ret = vpaudio_free_vmalloc_buffer(substream);
	if (ret < 0) {
		DBG("vpaudio_playback_hw_free can not free memory.\n");
	}

	return 0;
}

static int vpaudio_playback_prepare(struct snd_pcm_substream *substream)
{
	//struct snd_pcm_runtime 	*runtime = substream->runtime;
	//struct vpaudio_dev	*vpdev 	 = substream->private_data;

	DBG("vpaudio_playback_prepare.\n");

	return 0;
}

static int vpaudio_playback_trigger(struct snd_pcm_substream *substream, int cmd)
{
	//struct  vpaudio_dev *vpdev = substream->private_data;
	//int	status;

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		DBG("playback : vpaudio_trigger stop.\n");
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		DBG("playback : vpaudio_trigger stop.\n");
		break;
	case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
		DBG("playback : vpaudio_trigger push.\n");
		break;
	case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
		DBG("playback : vpaudio_trigger release.\n");
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static snd_pcm_uframes_t vpaudio_playback_pointer(struct snd_pcm_substream *substream)
{
	return 0;
}

static struct snd_pcm_ops vpaudio_capture_ops = {
	.open 		=	vpaudio_capture_open,
	.close 		=	vpaudio_capture_close,
	.ioctl 		=	snd_pcm_lib_ioctl,
	.hw_params 	=	vpaudio_capture_hw_params,
	.hw_free 	=	vpaudio_capture_hw_free,
	.prepare 	=	vpaudio_capture_prepare,
	.trigger 	=	vpaudio_capture_trigger,
	.pointer 	=	vpaudio_capture_pointer,
};

static struct snd_pcm_ops vpaudio_playback_ops = {
	.open 		=	vpaudio_playback_open,
	.close 		=	vpaudio_playback_close,
	.ioctl 		=	snd_pcm_lib_ioctl,
	.hw_params 	=	vpaudio_playback_hw_params,
	.hw_free 	=	vpaudio_playback_hw_free,
	.prepare 	=	vpaudio_playback_prepare,
	.trigger 	=	vpaudio_playback_trigger,
	.pointer 	=	vpaudio_playback_pointer,
};

static struct usb_device_id vpaudio_table[] = {
	{ USB_DEVICE(0x04FC, 0x0260) },
	{ }
};

MODULE_DEVICE_TABLE (usb, vpaudio_table);

static int vpaudio_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
	struct usb_device *udev = interface_to_usbdev(intf);

	const struct usb_host_interface 	*interface = NULL;
	const struct usb_endpoint_descriptor 	*endpoint = NULL;
	struct vpaudio_dev 			*vpdev = NULL;
	struct snd_pcm 				*pcm = NULL;

	int numas = 0, ias = 0, err;

	DBG("Vusial Presenter Audio Device Connected\n");

	numas = intf->num_altsetting;
	DBG("numas = %d.\n", numas);

	for (ias = 0; ias < numas; ias++) {
		interface = &intf->altsetting[ias];
		endpoint  = &interface->endpoint[1].desc;

		DBG("endpoint->bEndpointAddress 0x%x, endpoint->bmAttributes 0x%x",
			endpoint->bEndpointAddress, endpoint->bmAttributes);

		if (USB_ENDPOINT_XFER_INT != (endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)) {
			DBG("Device have not a INT endpoint!\n");
			return -ENODEV;
		}

		if (USB_DIR_IN != (endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)) {
			DBG("Device have not a IN endpoint!\n");
			return -ENODEV;
		}
	}
	
	if ((vpdev = kmalloc(sizeof(struct vpaudio_dev), GFP_KERNEL)) == NULL) {
		DBG("Could not allocate kernel memory for vpaudio_dev.\n");
		return -ENOMEM;
	}
	
	spin_lock_init(&vpdev->lock);
	
	vpdev->bInterval 	= endpoint->bInterval;
	vpdev->INTEndpoint 	= endpoint->bEndpointAddress;
	vpdev->pipe		= usb_rcvintpipe(udev, endpoint->bEndpointAddress);
	vpdev->INTMaxsize	= usb_maxpacket(udev, vpdev->pipe, usb_pipeout(vpdev->pipe));
	
	DBG("maxp = %d, bInterval = %d, pipe = %d.\n", vpdev->INTMaxsize, vpdev->bInterval, vpdev->pipe);

	vpdev->data = kmalloc(vpdev->INTMaxsize*sizeof(unsigned char), GFP_KERNEL);
	memset(vpdev->data, 0, vpdev->INTMaxsize*sizeof(unsigned char));
	vpdev->start = 0;
	vpdev->isready = 0;
	vpdev->hwptr_done = 0;
	if (!vpdev->data) {
		DBG("Could not allocate kernel memory for data.\n");
		return -ENOMEM;
	}

	vpdev->irq = usb_alloc_urb(0, GFP_KERNEL);
	if (!vpdev->irq) {
		DBG("Could not allocate kernel memory for irq.\n");
		return -ENOMEM;
	}

	usb_fill_int_urb(vpdev->irq, vpdev->udev, vpdev->pipe, vpdev->data,
					 vpdev->INTMaxsize, usb_audio_irq, vpdev, 1);
	vpdev->irq->dev = udev;
	/*status = usb_submit_urb(vpdev->irq, GFP_KERNEL);
	if (status) {
		DBG("irq can't resubmit intr.\n");
	}
	else {
		DBG("send urb ok.\n");
	}*/

	mutex_lock(&register_mutex);
	vpdev->card = snd_card_new(-1, "vpaudio", THIS_MODULE, 0);
	if (vpdev->card) {
		DBG("snd card create success!\n");
	}
	else {
		DBG("snd card create fail!\n");
		goto error;
	}

	err = snd_pcm_new(vpdev->card, "Visual Presenter Audio", 0, 1, 1, &pcm);
	if (err) {
		DBG("snd new pcm fail!\n");
		goto error;
	}
	else {
		DBG("snd new pcm success!\n");
	}
	pcm->private_data = vpdev;
	strcpy(pcm->name, "Visual Presenter Audio LV1");
	vpdev->pcm = pcm;

	snd_pcm_set_ops(vpdev->pcm, SNDRV_PCM_STREAM_CAPTURE, 
			&vpaudio_capture_ops);
	snd_pcm_set_ops(vpdev->pcm, SNDRV_PCM_STREAM_PLAYBACK, 
			&vpaudio_playback_ops);
	
	if (snd_card_register(vpdev->card) < 0) {
		DBG("snd card register fail!\n");
		goto error;
	}
	else {
		DBG("snd card register success!\n");
	}
	mutex_unlock(&register_mutex);

	usb_set_intfdata(intf, vpdev);

	DBG("Visual Presenter Audio Probe success!.\n");

	return 0;
error:
	usb_free_urb(vpdev->irq);
	kfree(vpdev->data);
	kfree(vpdev);
	return -ENOMEM;
}

static void vpaudio_disconnect(struct usb_interface *intf)
{
	struct vpaudio_dev *vpdev = usb_get_intfdata(intf);

	usb_set_intfdata(intf, NULL);
	
	mutex_lock(&register_mutex);
	snd_card_free(vpdev->card);
	DBG("Visual Presneter Audio free card!\n");
	//kfree(vpdev->pcm);
	DBG("Visual Presneter Audio free pcm!\n");
	//usb_kill_urb(vpdev->irq);
	DBG("Visual Presneter Audio kill irq!\n");
	usb_free_urb(vpdev->irq);
	DBG("Visual Presneter Audio free irq!\n");
	kfree(vpdev->data);
	kfree(vpdev);
	DBG("Visual Presneter Audio disconnect!\n");
	mutex_unlock(&register_mutex);
}

static struct usb_driver vpaudio_driver = {
	.name		= "Visual Presenter USB Audio Device",
	.probe		= vpaudio_probe,
	.disconnect 	= vpaudio_disconnect,
	.id_table	= vpaudio_table 
};

static int __init usb_vpaudio_init(void)
{
	int retval;
	DBG("Visual Presenter Audio driver startup.\n");
	retval = usb_register(&vpaudio_driver);
	if (retval) {
		DBG("usb_register fail!.\n");
	}
	
	return retval;
}

static void __exit usb_vpaudio_exit(void)
{
	DBG("Visual Presenter Audio driver shutdown.\n");
	usb_deregister(&vpaudio_driver);
}

module_init(usb_vpaudio_init);
module_exit(usb_vpaudio_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("zhangwei <zhang.wei.qd@gmail.com>");
MODULE_DESCRIPTION("Visual Presenter Audio Driver");
